/*
 *  pe_tile_rotation.c
 *  ProteinEncoder
 *
 *  Created by Naoto Morikawa on 6/20/06.
 *  Copyright (C) 2006, 2007, 2008 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 *  2007-02-15 version 0.3 release
 *  2007-04-15 version 0.4 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pe_main.h"
#include "pe_encoding.h"
#include "pe_print.h"

extern void print_trn_table(void);

extern struct pe_protein my_protein;
extern struct pe_coor_system my_coor;
extern struct pe_trn_table my_trn_table[];
extern struct pe_trn_table my_trn_table_save[];

void print_mtrx(struct pe_matrix * pt_M)
{
	fprintf(STDOUT, "print_mat() :               | %8.3f, %8.3f, %8.3f |\n",
		pt_M->a11, pt_M->a12, pt_M->a13);
	fprintf(STDOUT, "print_mat() :         Mat = | %8.3f, %8.3f, %8.3f |\n",
		pt_M->a21, pt_M->a22, pt_M->a23);
	fprintf(STDOUT, "print_mat() :               | %8.3f, %8.3f, %8.3f |\n",
		pt_M->a31, pt_M->a32, pt_M->a33);
	return;
}

void print_vec(struct pe_vector * pt_vec)
{
	fprintf(STDOUT, "print_vec() :         vec = ( %8.3f, %8.3f, %8.3f )\n",
		pt_vec->x, pt_vec->y, pt_vec->z);
	return;
}

/* c = a x b */
void cross(struct pe_vector * pt_c, struct pe_vector * pt_a,
	   struct pe_vector * pt_b)
{
	pt_c->x = pt_a->y * pt_b->z - pt_a->z * pt_b->y;
	pt_c->y = pt_a->z * pt_b->x - pt_a->x * pt_b->z;
	pt_c->z = pt_a->x * pt_b->y - pt_a->y * pt_b->x;
	return;
}

float determinant(struct pe_matrix * pt_mat)
{
	float det;
	det = pt_mat->a11 * (pt_mat->a22 * pt_mat->a33
			     - pt_mat->a23 * pt_mat->a32)
		- pt_mat->a12 * (pt_mat->a21 * pt_mat->a33
				 - pt_mat->a23 * pt_mat->a31)
		+ pt_mat->a13 * (pt_mat->a21 * pt_mat->a32
				 - pt_mat->a22 * pt_mat->a31);
	return det;
}

void inverse_mat(struct pe_matrix * pt_inv, struct pe_matrix * pt_mat)
{
	float detA11, detA21, detA31;
	float detA12, detA22, detA32;
	float detA13, detA23, detA33;
	float detA;

	detA = determinant(pt_mat);
	if (!FLOAT_NEQ(detA, 1.0)) {
		fprintf(STDERR, "inverse_mat() : ERROR  - |A| == %8.3f -\n",
			detA);
		print_mtrx(pt_mat);
		exit(1);
	}
	/* The first row */
	detA11 = pt_mat->a22 * pt_mat->a33 - pt_mat->a23 * pt_mat->a32;
	pt_inv->a11 = detA11 / detA;
	detA21 = pt_mat->a12 * pt_mat->a33 - pt_mat->a13 * pt_mat->a32;
	pt_inv->a12 = -1 * detA21 / detA;
	detA31 = pt_mat->a12 * pt_mat->a23 - pt_mat->a13 * pt_mat->a22;
	pt_inv->a13 = detA31 / detA;
	/* The second row */
	detA12 = pt_mat->a21 * pt_mat->a33 - pt_mat->a23 * pt_mat->a31;
	pt_inv->a21 = -1 * detA12 / detA;
	detA22 = pt_mat->a11 * pt_mat->a33 - pt_mat->a13 * pt_mat->a31;
	pt_inv->a22 = detA22 / detA;
	detA32 = pt_mat->a11 * pt_mat->a23 - pt_mat->a13 * pt_mat->a21;
	pt_inv->a23 = -1 * detA32 / detA;
	/* The third row */
	detA13 = pt_mat->a21 * pt_mat->a32 - pt_mat->a22 * pt_mat->a31;
	pt_inv->a31 = detA13 / detA;
	detA23 = pt_mat->a11 * pt_mat->a32 - pt_mat->a12 * pt_mat->a31;
	pt_inv->a32 = -1 * detA23 / detA;
	detA33 = pt_mat->a11 * pt_mat->a22 - pt_mat->a12 * pt_mat->a21;
	pt_inv->a33 = detA33 / detA;
	return;
}

void mat_mul(struct pe_matrix * pt_C, struct pe_matrix * pt_A,
	     struct pe_matrix * pt_B)
{				/* C = AB */
	/* The first row */
	pt_C->a11 = pt_A->a11 * pt_B->a11 + pt_A->a12 * pt_B->a21
		+ pt_A->a13 * pt_B->a31;
	pt_C->a12 = pt_A->a11 * pt_B->a12 + pt_A->a12 * pt_B->a22
		+ pt_A->a13 * pt_B->a32;
	pt_C->a13 = pt_A->a11 * pt_B->a13 + pt_A->a12 * pt_B->a23
		+ pt_A->a13 * pt_B->a33;
	/* The second row */
	pt_C->a21 = pt_A->a21 * pt_B->a11 + pt_A->a22 * pt_B->a21
		+ pt_A->a23 * pt_B->a31;
	pt_C->a22 = pt_A->a21 * pt_B->a12 + pt_A->a22 * pt_B->a22
		+ pt_A->a23 * pt_B->a32;
	pt_C->a23 = pt_A->a21 * pt_B->a13 + pt_A->a22 * pt_B->a23
		+ pt_A->a23 * pt_B->a33;
	/* The third row */
	pt_C->a31 = pt_A->a31 * pt_B->a11 + pt_A->a32 * pt_B->a21
		+ pt_A->a33 * pt_B->a31;
	pt_C->a32 = pt_A->a31 * pt_B->a12 + pt_A->a32 * pt_B->a22
		+ pt_A->a33 * pt_B->a32;
	pt_C->a33 = pt_A->a31 * pt_B->a13 + pt_A->a32 * pt_B->a23
		+ pt_A->a33 * pt_B->a33;
	return;
}

int chk_eq(struct pe_matrix * pt_A, struct pe_matrix * pt_B)
{				/* return 1 if A == B, else return 0 */
	if (FLOAT_NEQ((pt_A->a11 - pt_B->a11), my_coor.ave_dis))
		return FAIL;
	if (FLOAT_NEQ((pt_A->a12 - pt_B->a12), my_coor.ave_dis))
		return FAIL;
	if (FLOAT_NEQ((pt_A->a13 - pt_B->a13), my_coor.ave_dis))
		return FAIL;

	if (FLOAT_NEQ((pt_A->a21 - pt_B->a21), my_coor.ave_dis))
		return FAIL;
	if (FLOAT_NEQ((pt_A->a22 - pt_B->a22), my_coor.ave_dis))
		return FAIL;
	if (FLOAT_NEQ((pt_A->a23 - pt_B->a23), my_coor.ave_dis))
		return FAIL;

	if (FLOAT_NEQ((pt_A->a31 - pt_B->a31), my_coor.ave_dis))
		return FAIL;
	if (FLOAT_NEQ((pt_A->a32 - pt_B->a32), my_coor.ave_dis))
		return FAIL;
	if (FLOAT_NEQ((pt_A->a33 - pt_B->a33), my_coor.ave_dis))
		return FAIL;

	return SUCCESS;
}

int set_vector(struct pe_vector * pvec, float val_x, float val_y, float val_z)
{
	float wrk_length;
	
	wrk_length = sqrt(val_x * val_x + val_y * val_y + val_z * val_z);
	if (!FLOAT_NEQ(wrk_length, 1.0)) {
		if (0) fprintf(STDERR, "set_vector() : DEBUG  |(%8.3f, %8.3f, %8.3f)| == %8.3f\n",
			val_x, val_y, val_z, wrk_length);
		return FAIL;
	}
	pvec->x = val_x / wrk_length;
	pvec->y = val_y / wrk_length;
	pvec->z = val_z / wrk_length;
	pvec->length = 1.0;
	return SUCCESS;
}

/* MM(cur_*) == new_* */
void rotation_matrix(struct pe_matrix * pt_M, float cur_x, float cur_y,
		     float cur_z, float new_x, float new_y, float new_z,
		     int cur_type)
{
	struct pe_vector vec_a, vec_b, vec_c, vec_ca, vec_cb;
	struct pe_matrix S, IS, E;
	
	/* a = current direction */
	if (set_vector(&vec_a, cur_x, cur_y, cur_z) == FAIL) {
		fprintf(STDERR, "rotation_matrix() : ERROR vec_a\n");
		exit(1);
	}
	/* b = target direction */
	if (set_vector(&vec_b, new_x, new_y, new_z) == FAIL) {
		fprintf(STDERR, "rotation_matrix() : ERROR vec_b\n");
		exit(1);
	}
	/* c = a x b */
	if (set_vector(&vec_c, vec_a.y * vec_b.z - vec_a.z * vec_b.y,
		       vec_a.z * vec_b.x - vec_a.x * vec_b.z,
		       vec_a.x * vec_b.y - vec_a.y * vec_b.x) == FAIL) {
		if (0) {
			fprintf(STDERR,
				"rotation_matrix() : DEBUG |vec_c| = |vec_a x vec_b| == 0.0\n");
			fprintf(STDERR,
				"rotation_matrix() :       vec_a = (%8.3f, %8.3f, %8.3f)\n",
				vec_a.x, vec_a.y, vec_a.z);
			fprintf(STDERR,
				"rotation_matrix() :       vec_b = (%8.3f, %8.3f, %8.3f)\n",
				vec_b.x, vec_b.y, vec_b.z);
		}
		if (!FLOAT_NEQ(vec_a.x - vec_b.x, 1.0)) {	/* if a == b */
			pt_M->a11 = 1.0; pt_M->a12 = 0.0; pt_M->a13 = 0.0;
			pt_M->a21 = 0.0; pt_M->a22 = 1.0; pt_M->a23 = 0.0;
			pt_M->a31 = 0.0; pt_M->a32 = 0.0; pt_M->a33 = 1.0;
			return;
		} else {
			set_vector(&vec_c, my_trn_table_save[cur_type].bx / 2.0
				   - my_trn_table_save[cur_type].cx,
				   my_trn_table_save[cur_type].by / 2.0
				   - my_trn_table_save[cur_type].cy,
				   my_trn_table_save[cur_type].bz / 2.0
				   - my_trn_table_save[cur_type].cz);
		}
	}
	/* ca = c x a */
	if (set_vector(&vec_ca, vec_c.y * vec_a.z - vec_c.z * vec_a.y,
		   vec_c.z * vec_a.x - vec_c.x * vec_a.z,
		       vec_c.x * vec_a.y - vec_c.y * vec_a.x) == FAIL) {
		fprintf(STDERR, "rotation_matrix() : ERROR vec_ca\n");
		fprintf(STDERR, "rotation_matrix() :       vec_c = (%8.3f, %8.3f, %8.3f)\n",
			vec_c.x, vec_c.y, vec_c.z);
		fprintf(STDERR, "rotation_matrix() :       vec_a = (%8.3f, %8.3f, %8.3f)\n",
			vec_a.x, vec_a.y, vec_a.z);
		exit(1);
	}
	/* cb = c x b */
	if (set_vector(&vec_cb, vec_c.y * vec_b.z - vec_c.z * vec_b.y,
		   vec_c.z * vec_b.x - vec_c.x * vec_b.z,
		       vec_c.x * vec_b.y - vec_c.y * vec_b.x) == FAIL) {
		fprintf(STDERR, "rotation_matrix() : ERROR vec_cb\n");
		fprintf(STDERR, "rotation_matrix() :       vec_c = (%8.3f, %8.3f, %8.3f)\n",
			vec_c.x, vec_c.y, vec_c.z);
		fprintf(STDERR, "rotation_matrix() :       vec_b = (%8.3f, %8.3f, %8.3f)\n",
			vec_b.x, vec_b.y, vec_b.z);
		exit(1);
	}

	/* M(vec_a, vec_ca, vec_c) = (vec_b, vec_cb, vec_c) */
	/* Set matrix S = (vec_a, vec_ca, vec_c) */
	S.a11 = vec_a.x;
	S.a21 = vec_a.y;
	S.a31 = vec_a.z;
	S.a12 = vec_ca.x;
	S.a22 = vec_ca.y;
	S.a32 = vec_ca.z;
	S.a13 = vec_c.x;
	S.a23 = vec_c.y;
	S.a33 = vec_c.z;
	/* Set matrix E = (vec_b, vec_cb, vec_c) */
	E.a11 = vec_b.x;
	E.a21 = vec_b.y;
	E.a31 = vec_b.z;
	E.a12 = vec_cb.x;
	E.a22 = vec_cb.y;
	E.a32 = vec_cb.z;
	E.a13 = vec_c.x;
	E.a23 = vec_c.y;
	E.a33 = vec_c.z;
	
	/* M = E*IS */
	if (0) {
		fprintf(STDERR, "rotation_matrix() : S is\n");
		print_mtrx(&S);
	}
	inverse_mat(&IS, &S);
	mat_mul(pt_M, &E, &IS);
	if (DEBUG) {	/* M*S <=> E */
		struct pe_matrix U;
		float wrk_error;
		
		mat_mul(&U, pt_M, &S);
		wrk_error = (U.a11 - E.a11) * (U.a11 - E.a11);
		wrk_error = (U.a21 - E.a21) * (U.a21 - E.a21) + wrk_error;
		wrk_error = (U.a31 - E.a31) * (U.a31 - E.a31) + wrk_error;
		wrk_error = (U.a12 - E.a12) * (U.a12 - E.a12) + wrk_error;
		wrk_error = (U.a22 - E.a22) * (U.a22 - E.a22) + wrk_error;
		wrk_error = (U.a32 - E.a32) * (U.a32 - E.a32) + wrk_error;
		wrk_error = (U.a13 - E.a13) * (U.a13 - E.a13) + wrk_error;
		wrk_error = (U.a23 - E.a23) * (U.a23 - E.a23) + wrk_error;
		wrk_error = (U.a33 - E.a33) * (U.a33 - E.a33) + wrk_error;
		if (FLOAT_NEQ(wrk_error, my_coor.ave_dis)) {
			fprintf(STDERR, "rotation_matrix() : ERROR  diff(M*S - E)= %f <==> 0 \n",
				wrk_error);
			exit(1);
		}
	}
	if (0) {
		fprintf(STDERR, "rotation_matrix() : Rotation matrix is\n");
		print_mtrx(pt_M);
	}
	return;
}

/* rot_mat(current tile) == (corresponding trn_table entry) */
void rotation_matrix2(struct pe_matrix * pt_M,  int t_ind,
		      struct pe_slt_tile * slt_tiles)
{
	struct pe_vector tile_top, tile_bot, tile_c;
	struct pe_vector entry_top, entry_bot, entry_c;
	float wrk_x, wrk_y, wrk_z;
	int cur_type;
	struct pe_matrix T, IT, E;
	
	/* tile_top = (top - center) of the current tile */
	wrk_x = slt_tiles[t_ind].tx - slt_tiles[t_ind].cx;
	wrk_y = slt_tiles[t_ind].ty - slt_tiles[t_ind].cy;
	wrk_z = slt_tiles[t_ind].tz - slt_tiles[t_ind].cz;
	if (set_vector(&tile_top, wrk_x, wrk_y, wrk_z) == FAIL) {
		fprintf(STDERR, "rotation_matrix() : ERROR tile_top\n");
		exit(1);
	}
	/* tile_bot = (bottom - center) of the current tile */
	wrk_x = slt_tiles[t_ind].bx - slt_tiles[t_ind].cx;
	wrk_y = slt_tiles[t_ind].by - slt_tiles[t_ind].cy;
	wrk_z = slt_tiles[t_ind].bz - slt_tiles[t_ind].cz;
	if (set_vector(&tile_bot, wrk_x, wrk_y, wrk_z) == FAIL) {
		fprintf(STDERR, "rotation_matrix() : ERROR tile_bot\n");
		exit(1);
	}
	/* tile_c = tile_top x tile_bot */
	if (set_vector(&tile_c, tile_top.y * tile_bot.z - tile_top.z * tile_bot.y,
		   tile_top.z * tile_bot.x - tile_top.x * tile_bot.z,
		       tile_top.x * tile_bot.y - tile_top.y * tile_bot.x) == FAIL) {
		fprintf(STDERR, "rotation_matrix() : ERROR tilr_c\n");
		exit(1);
	}
	/* Set matrix T = (top, bot, c) */
	T.a11 = tile_top.x;
	T.a21 = tile_top.y;
	T.a31 = tile_top.z;
	T.a12 = tile_bot.x;
	T.a22 = tile_bot.y;
	T.a32 = tile_bot.z;
	T.a13 = tile_c.x;
	T.a23 = tile_c.y;
	T.a33 = tile_c.z;
	if (0) {
		fprintf(STDERR, "rotation_matrix2() : T is\n");
		print_mtrx(&T);
	}
	inverse_mat(&IT, &T);
	
	cur_type = slt_tiles[t_ind].type;
	/* entry_top = (top - center) of the trn_table entry */
	wrk_x = - my_trn_table[cur_type].cx;
	wrk_y = - my_trn_table[cur_type].cy;
	wrk_z = - my_trn_table[cur_type].cz;
	if (set_vector(&entry_top, wrk_x, wrk_y, wrk_z) == FAIL) {
		fprintf(STDERR, "rotation_matrix() : ERROR entry_top\n");
		exit(1);
	}
	/* entry_top = (bottom - center) of the trn_table entry */
	wrk_x = my_trn_table[cur_type].bx - my_trn_table[cur_type].cx;
	wrk_y = my_trn_table[cur_type].by - my_trn_table[cur_type].cy;
	wrk_z = my_trn_table[cur_type].bz - my_trn_table[cur_type].cz;
	if (set_vector(&entry_bot, wrk_x, wrk_y, wrk_z) == FAIL) {
		fprintf(STDERR, "rotation_matrix() : ERROR entry_bot\n");
		exit(1);
	}
	/* entry_c = entry_top x entry_bot */
	if (set_vector(&entry_c,
		       entry_top.y * entry_bot.z - entry_top.z * entry_bot.y,
		       entry_top.z * entry_bot.x - entry_top.x * entry_bot.z,
		       entry_top.x * entry_bot.y - entry_top.y * entry_bot.x)
	    == FAIL) {
		fprintf(STDERR, "rotation_matrix() : ERROR entry_c\n");
		exit(1);
	}
	/* Set matrix E = (top, bot, c) */
	E.a11 = entry_top.x;
	E.a21 = entry_top.y;
	E.a31 = entry_top.z;
	E.a12 = entry_bot.x;
	E.a22 = entry_bot.y;
	E.a32 = entry_bot.z;
	E.a13 = entry_c.x;
	E.a23 = entry_c.y;
	E.a33 = entry_c.z;
	
	/* Since M*T = E, M = E*IV */
	if (DEBUG) {
		float wrk_det;
		
		wrk_det = determinant(&T);
		if (!FLOAT_NEQ(wrk_det, my_coor.ave_dis)) {
			fprintf(STDERR, "rotation_matrix2() : ERROR  |T|= %f\n",
				wrk_det);
			exit(1);
		}
	}
	inverse_mat(&IT, &T);
	mat_mul(pt_M, &E, &IT);		/* M = R*T */
	if (DEBUG) {
		struct pe_matrix U;
		float wrk_error;
		
		mat_mul(&U, pt_M, &T);
		wrk_error = (U.a11 - E.a11) * (U.a11 - E.a11);
		wrk_error = (U.a21 - E.a21) * (U.a21 - E.a21) + wrk_error;
		wrk_error = (U.a31 - E.a31) * (U.a31 - E.a31) + wrk_error;
		wrk_error = (U.a12 - E.a12) * (U.a12 - E.a12) + wrk_error;
		wrk_error = (U.a22 - E.a22) * (U.a22 - E.a22) + wrk_error;
		wrk_error = (U.a32 - E.a32) * (U.a32 - E.a32) + wrk_error;
		wrk_error = (U.a13 - E.a13) * (U.a13 - E.a13) + wrk_error;
		wrk_error = (U.a23 - E.a23) * (U.a23 - E.a23) + wrk_error;
		wrk_error = (U.a33 - E.a33) * (U.a33 - E.a33) + wrk_error;
		if (FLOAT_NEQ(wrk_error, my_coor.ave_dis)) {
			fprintf(STDERR, "rotation_matrix2() : ERROR  diff(M*T - E)= %f <==> 0 \n",
				wrk_error);
			exit(1);
		}
	}
	return;
}

#define FIRST_ROW(pA, wrk_x, wrk_y, wrk_z) ((pA)->a11 * (wrk_x) \
				+ (pA)->a12 * (wrk_y) + (pA)->a13 * (wrk_z))
#define SECOND_ROW(pA, wrk_x, wrk_y, wrk_z) ((pA)->a21 * (wrk_x) \
				+ (pA)->a22 * (wrk_y) + (pA)->a23 * (wrk_z))
#define THIRD_ROW(pA, wrk_x, wrk_y, wrk_z) ((pA)->a31 * (wrk_x) \
				+ (pA)->a32 * (wrk_y) + (pA)->a33 * (wrk_z))

void compute_new_trn_top(struct pe_matrix * p_mat, int t_ind,
		    struct pe_slt_tile * slt_tiles, float * ptop_x,
		    float * ptop_y, float * ptop_z)
{
	int cur_type;
	float wrk_x, wrk_y, wrk_z;
	float cen_x, cen_y, cen_z;
	
	cur_type = slt_tiles[t_ind].type;
	
	/* center of the rotation */
	cen_x = my_trn_table_save[cur_type].cx;
	cen_y = my_trn_table_save[cur_type].cy;
	cen_z = my_trn_table_save[cur_type].cz;
	
	/* New top */
	wrk_x = - cen_x;
	wrk_y = - cen_y;
	wrk_z = - cen_z;
	*ptop_x = FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x;
	*ptop_y = SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y;
	*ptop_z = THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z;
	
	/* New center */
	my_trn_table[cur_type].cx = cen_x - *ptop_x;
	my_trn_table[cur_type].cy = cen_y - *ptop_y;
	my_trn_table[cur_type].cz = cen_z - *ptop_z;
	
	/* New bottom */
	wrk_x = my_trn_table_save[cur_type].bx - cen_x;
	wrk_y = my_trn_table_save[cur_type].by - cen_y;
	wrk_z = my_trn_table_save[cur_type].bz - cen_z;
	my_trn_table[cur_type].bx = FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z)
		+ cen_x - *ptop_x;
	my_trn_table[cur_type].by = SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z)
		+ cen_y - *ptop_y;
	my_trn_table[cur_type].bz = THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z)
		+ cen_z - *ptop_z;
	return;
}

void compute_new_trn_downward(struct pe_matrix * p_mat, int t_ind,
			      struct pe_slt_tile * slt_tiles, float top_x,
			      float top_y, float top_z)
{
	int cur_type;
	float wrk_x, wrk_y, wrk_z;
	float cen_x, cen_y, cen_z;
	
	cur_type = slt_tiles[t_ind].type;
	/* center of the rotation */
	cen_x = my_trn_table_save[cur_type].cx;
	cen_y = my_trn_table_save[cur_type].cy;
	cen_z = my_trn_table_save[cur_type].cz;

	wrk_x = my_trn_table_save[cur_type].downward_cx - cen_x;
	wrk_y = my_trn_table_save[cur_type].downward_cy - cen_y;
	wrk_z = my_trn_table_save[cur_type].downward_cz - cen_z;
	my_trn_table[cur_type].downward_cx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x - top_x;
	my_trn_table[cur_type].downward_cy
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].downward_cz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	if (0) {
		float wrk_length, wrk2_length;
		
		wrk_x = my_trn_table_save[cur_type].downward_cx
			- my_trn_table_save[cur_type].cx;
		wrk_y = my_trn_table_save[cur_type].downward_cy
			- my_trn_table_save[cur_type].cy;
		wrk_z = my_trn_table_save[cur_type].downward_cz
			- my_trn_table_save[cur_type].cz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		wrk_x = my_trn_table[cur_type].downward_cx
			- my_trn_table[cur_type].cx;
		wrk_y = my_trn_table[cur_type].downward_cy
			- my_trn_table[cur_type].cy;
		wrk_z = my_trn_table[cur_type].downward_cz
			- my_trn_table[cur_type].cz;
		wrk2_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				   + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk2_length, 1.0)) {
			fprintf(STDERR,
				"compute_new_trn_downward() : ERROR wrk_length= %f <=> wrk2_length= %f\n",
				wrk_length, wrk2_length);
			exit(1);
		}
	}
	
	wrk_x = my_trn_table_save[cur_type].downward_U_next_cx - cen_x;
	wrk_y = my_trn_table_save[cur_type].downward_U_next_cy - cen_y;
	wrk_z = my_trn_table_save[cur_type].downward_U_next_cz - cen_z;
	my_trn_table[cur_type].downward_U_next_cx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x - top_x;
	my_trn_table[cur_type].downward_U_next_cy
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].downward_U_next_cz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	
	wrk_x = my_trn_table_save[cur_type].downward_U_tx - cen_x;
	wrk_y = my_trn_table_save[cur_type].downward_U_ty - cen_y;
	wrk_z = my_trn_table_save[cur_type].downward_U_tz - cen_z;
	my_trn_table[cur_type].downward_U_tx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x - top_x;
	my_trn_table[cur_type].downward_U_ty
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].downward_U_tz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	
	wrk_x = my_trn_table_save[cur_type].downward_U_bx - cen_x;
	wrk_y = my_trn_table_save[cur_type].downward_U_by - cen_y;
	wrk_z = my_trn_table_save[cur_type].downward_U_bz - cen_z;
	my_trn_table[cur_type].downward_U_bx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x - top_x;
	my_trn_table[cur_type].downward_U_by
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].downward_U_bz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	if (0) {
		float wrk_length, wrk_length2;
		
		wrk_x = my_trn_table[cur_type].downward_U_bx;
		wrk_y = my_trn_table[cur_type].downward_U_by;
		wrk_z = my_trn_table[cur_type].downward_U_bz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		if (!FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"compute_new_trn_downward() : ERROR my_trn_table[%d].downward_U_b*= (%8.3f, %8.3f, %8.3f)\n",
				cur_type, my_trn_table[cur_type].downward_U_bx, 
				my_trn_table[cur_type].downward_U_by,
				my_trn_table[cur_type].downward_U_bz);
			exit(1);
		}
		
		wrk_x = my_trn_table[cur_type].downward_cx - my_trn_table[cur_type].downward_U_tx;
		wrk_y = my_trn_table[cur_type].downward_cy - my_trn_table[cur_type].downward_U_ty;
		wrk_z = my_trn_table[cur_type].downward_cz - my_trn_table[cur_type].downward_U_tz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		wrk_x = my_trn_table[cur_type].downward_cx - my_trn_table[cur_type].downward_U_bx;
		wrk_y = my_trn_table[cur_type].downward_cy - my_trn_table[cur_type].downward_U_by;
		wrk_z = my_trn_table[cur_type].downward_cz - my_trn_table[cur_type].downward_U_bz;
		wrk_length2 = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				   + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk_length2, 1.0)) {
			fprintf(STDERR,
				"compute_new_trn_downward() : DEBUG my_trn_table[%d].downward_U |cen - top|= %f\n",
				cur_type, wrk_length);
			fprintf(STDERR,
				"compute_new_trn_downward() : DEBUG my_trn_table[%d].downward_U |cen - bot|= %f\n",
				cur_type, wrk_length2);
			exit(1);
		}
	}
	
	wrk_x = my_trn_table_save[cur_type].downward_D_next_cx - cen_x;
	wrk_y = my_trn_table_save[cur_type].downward_D_next_cy - cen_y;
	wrk_z = my_trn_table_save[cur_type].downward_D_next_cz - cen_z;
	my_trn_table[cur_type].downward_D_next_cx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x - top_x;
	my_trn_table[cur_type].downward_D_next_cy
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].downward_D_next_cz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	
	wrk_x = my_trn_table_save[cur_type].downward_D_tx - cen_x;
	wrk_y = my_trn_table_save[cur_type].downward_D_ty - cen_y;
	wrk_z = my_trn_table_save[cur_type].downward_D_tz - cen_z;
	my_trn_table[cur_type].downward_D_tx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x - top_x;
	my_trn_table[cur_type].downward_D_ty
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].downward_D_tz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	
	wrk_x = my_trn_table_save[cur_type].downward_D_bx - cen_x;
	wrk_y = my_trn_table_save[cur_type].downward_D_by - cen_y;
	wrk_z = my_trn_table_save[cur_type].downward_D_bz - cen_z;
	my_trn_table[cur_type].downward_D_bx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x - top_x;
	my_trn_table[cur_type].downward_D_by
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].downward_D_bz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	if (0) {
		float wrk_length, wrk_length2;
		
		wrk_x = my_trn_table[cur_type].downward_D_bx;
		wrk_y = my_trn_table[cur_type].downward_D_by;
		wrk_z = my_trn_table[cur_type].downward_D_bz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		if (!FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"compute_new_trn_downward() : ERROR my_trn_table[%d].downward_D_b*= (%8.3f, %8.3f, %8.3f)\n",
				cur_type, my_trn_table[cur_type].downward_D_bx, 
				my_trn_table[cur_type].downward_D_by,
				my_trn_table[cur_type].downward_D_bz);
			exit(1);
		}
		wrk_x = my_trn_table[cur_type].downward_cx - my_trn_table[cur_type].downward_D_tx;
		wrk_y = my_trn_table[cur_type].downward_cy - my_trn_table[cur_type].downward_D_ty;
		wrk_z = my_trn_table[cur_type].downward_cz - my_trn_table[cur_type].downward_D_tz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		wrk_x = my_trn_table[cur_type].downward_cx - my_trn_table[cur_type].downward_D_bx;
		wrk_y = my_trn_table[cur_type].downward_cy - my_trn_table[cur_type].downward_D_by;
		wrk_z = my_trn_table[cur_type].downward_cz - my_trn_table[cur_type].downward_D_bz;
		wrk_length2 = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				   + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk_length2, 1.0)) {
			fprintf(STDERR,
				"compute_new_trn_downward() : DEBUG my_trn_table[%d].downward_D |cen - top|= %f\n",
				cur_type, wrk_length);
			fprintf(STDERR,
				"compute_new_trn_downward() : DEBUG my_trn_table[%d].downward_D |cen - bot|= %f\n",
				cur_type, wrk_length2);
			exit(1);
		}
	}
	return;
}

void compute_new_trn_upward(struct pe_matrix * p_mat, int t_ind,
			      struct pe_slt_tile * slt_tiles, float top_x,
			      float top_y, float top_z)
{
	int cur_type;
	float wrk_x, wrk_y, wrk_z;
	float cen_x, cen_y, cen_z;
	
	cur_type = slt_tiles[t_ind].type;
	/* center of the rotation */
	cen_x = my_trn_table_save[cur_type].cx;
	cen_y = my_trn_table_save[cur_type].cy;
	cen_z = my_trn_table_save[cur_type].cz;
	
	wrk_x = my_trn_table_save[cur_type].upward_cx - cen_x;
	wrk_y = my_trn_table_save[cur_type].upward_cy - cen_y;
	wrk_z = my_trn_table_save[cur_type].upward_cz - cen_z;
	my_trn_table[cur_type].upward_cx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x - top_x;
	my_trn_table[cur_type].upward_cy
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].upward_cz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	if (0) {
		float wrk_length, wrk2_length;
		
		wrk_x = my_trn_table_save[cur_type].upward_cx
			- my_trn_table_save[cur_type].cx;
		wrk_y = my_trn_table_save[cur_type].upward_cy
			- my_trn_table_save[cur_type].cy;
		wrk_z = my_trn_table_save[cur_type].upward_cz
			- my_trn_table_save[cur_type].cz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		wrk_x = my_trn_table[cur_type].upward_cx
			- my_trn_table[cur_type].cx;
		wrk_y = my_trn_table[cur_type].upward_cy
			- my_trn_table[cur_type].cy;
		wrk_z = my_trn_table[cur_type].upward_cz
			- my_trn_table[cur_type].cz;
		wrk2_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				   + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk2_length, 1.0)) {
			fprintf(STDERR,
				"compute_new_trn_upward() : ERROR wrk_length= %f <=> wrk2_length= %f\n",
				wrk_length, wrk2_length);
			exit(1);
		}
	}
	
	wrk_x = my_trn_table_save[cur_type].upward_U_next_cx - cen_x;
	wrk_y = my_trn_table_save[cur_type].upward_U_next_cy - cen_y;
	wrk_z = my_trn_table_save[cur_type].upward_U_next_cz - cen_z;
	my_trn_table[cur_type].upward_U_next_cx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x - top_x;
	my_trn_table[cur_type].upward_U_next_cy
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].upward_U_next_cz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	
	wrk_x = my_trn_table_save[cur_type].upward_U_tx - cen_x;
	wrk_y = my_trn_table_save[cur_type].upward_U_ty - cen_y;
	wrk_z = my_trn_table_save[cur_type].upward_U_tz - cen_z;
	my_trn_table[cur_type].upward_U_tx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x - top_x;
	my_trn_table[cur_type].upward_U_ty
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].upward_U_tz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	
	wrk_x = my_trn_table_save[cur_type].upward_U_bx - cen_x;
	wrk_y = my_trn_table_save[cur_type].upward_U_by - cen_y;
	wrk_z = my_trn_table_save[cur_type].upward_U_bz - cen_z;
	my_trn_table[cur_type].upward_U_bx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x - top_x;
	my_trn_table[cur_type].upward_U_by
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].upward_U_bz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	if (0) {
		float wrk_length, wrk_length2;
		
		wrk_x = my_trn_table[cur_type].upward_U_bx;
		wrk_y = my_trn_table[cur_type].upward_U_by;
		wrk_z = my_trn_table[cur_type].upward_U_bz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		if (!FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"compute_new_trn_upward() : ERROR my_trn_table[%d].upward_U_b*= (%8.3f, %8.3f, %8.3f)\n",
				cur_type, my_trn_table[cur_type].upward_U_bx, 
				my_trn_table[cur_type].upward_U_by,
				my_trn_table[cur_type].upward_U_bz);
			exit(1);
		}
		
		wrk_x = my_trn_table[cur_type].upward_cx - my_trn_table[cur_type].upward_U_tx;
		wrk_y = my_trn_table[cur_type].upward_cy - my_trn_table[cur_type].upward_U_ty;
		wrk_z = my_trn_table[cur_type].upward_cz - my_trn_table[cur_type].upward_U_tz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		wrk_x = my_trn_table[cur_type].upward_cx - my_trn_table[cur_type].upward_U_bx;
		wrk_y = my_trn_table[cur_type].upward_cy - my_trn_table[cur_type].upward_U_by;
		wrk_z = my_trn_table[cur_type].upward_cz - my_trn_table[cur_type].upward_U_bz;
		wrk_length2 = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				   + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk_length2, 1.0)) {
			fprintf(STDERR,
				"compute_new_trn_upward() : DEBUG my_trn_table[%d].upward_U |cen - top|= %f\n",
				cur_type, wrk_length);
			fprintf(STDERR,
				"compute_new_trn_upward() : DEBUG my_trn_table[%d].upward_U |cen - bot|= %f\n",
				cur_type, wrk_length2);
			exit(1);
		}
	}
	
	wrk_x = my_trn_table_save[cur_type].upward_D_next_cx - cen_x;
	wrk_y = my_trn_table_save[cur_type].upward_D_next_cy - cen_y;
	wrk_z = my_trn_table_save[cur_type].upward_D_next_cz - cen_z;
	my_trn_table[cur_type].upward_D_next_cx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x - top_x;
	my_trn_table[cur_type].upward_D_next_cy
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].upward_D_next_cz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	
	wrk_x = my_trn_table_save[cur_type].upward_D_tx - cen_x;
	wrk_y = my_trn_table_save[cur_type].upward_D_ty - cen_y;
	wrk_z = my_trn_table_save[cur_type].upward_D_tz - cen_z;
	my_trn_table[cur_type].upward_D_tx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x  - top_x;
	my_trn_table[cur_type].upward_D_ty
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].upward_D_tz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	
	wrk_x = my_trn_table_save[cur_type].upward_D_bx - cen_x;
	wrk_y = my_trn_table_save[cur_type].upward_D_by - cen_y;
	wrk_z = my_trn_table_save[cur_type].upward_D_bz - cen_z;
	my_trn_table[cur_type].upward_D_bx
		= FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x  - top_x;
	my_trn_table[cur_type].upward_D_by
		= SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y - top_y;
	my_trn_table[cur_type].upward_D_bz
		= THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z - top_z;
	if (0) {
		float wrk_length, wrk_length2;
		
		wrk_x = my_trn_table[cur_type].upward_D_bx;
		wrk_y = my_trn_table[cur_type].upward_D_by;
		wrk_z = my_trn_table[cur_type].upward_D_bz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		if (!FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"compute_new_trn_upward() : ERROR my_trn_table[%d].upward_D_b*= (%8.3f, %8.3f, %8.3f)\n",
				cur_type, my_trn_table[cur_type].upward_D_bx, 
				my_trn_table[cur_type].upward_D_by,
				my_trn_table[cur_type].upward_D_bz);
			exit(1);
		}
		
		wrk_x = my_trn_table[cur_type].upward_cx - my_trn_table[cur_type].upward_D_tx;
		wrk_y = my_trn_table[cur_type].upward_cy - my_trn_table[cur_type].upward_D_ty;
		wrk_z = my_trn_table[cur_type].upward_cz - my_trn_table[cur_type].upward_D_tz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		wrk_x = my_trn_table[cur_type].upward_cx - my_trn_table[cur_type].upward_D_bx;
		wrk_y = my_trn_table[cur_type].upward_cy - my_trn_table[cur_type].upward_D_by;
		wrk_z = my_trn_table[cur_type].upward_cz - my_trn_table[cur_type].upward_D_bz;
		wrk_length2 = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				   + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk_length2, 1.0)) {
			fprintf(STDERR,
				"compute_new_trn_upward() : DEBUG my_trn_table[%d].upward_D |cen - top|= %f\n",
				cur_type, wrk_length);
			fprintf(STDERR,
				"compute_new_trn_upward() : DEBUG my_trn_table[%d].upward_D |cen - bot|= %f\n",
				cur_type, wrk_length2);
			exit(1);
		}
	}
	return;
}

void compute_new_tile_top(struct pe_matrix * p_mat, int t_ind,
			 struct pe_slt_tile * slt_tiles)
{
	float wrk_x, wrk_y, wrk_z;
	float cen_x, cen_y, cen_z;
	
	cen_x = slt_tiles[t_ind].cx;
	cen_y = slt_tiles[t_ind].cy;
	cen_z = slt_tiles[t_ind].cz;
	
	/* New top */
	wrk_x = slt_tiles[t_ind].tx - cen_x;
	wrk_y = slt_tiles[t_ind].ty - cen_y;
	wrk_z = slt_tiles[t_ind].tz - cen_z;
	slt_tiles[t_ind].tx = FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x;
	slt_tiles[t_ind].ty = SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y;
	slt_tiles[t_ind].tz = THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z;
	if (0) {
		float wrk_length, wrk2_length;
		
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		wrk_x = slt_tiles[t_ind].tx - cen_x;
		wrk_y = slt_tiles[t_ind].ty - cen_y;
		wrk_z = slt_tiles[t_ind].tz - cen_z;
		wrk2_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				   + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk2_length, 1.0)) {
			fprintf(STDERR,
				"compute_new_tile_top() : ERROR wrk_length= %f <=> wrk2_length= %f\n",
				wrk_length, wrk2_length);
			exit(1);
		}
	}
	
	/* New bottom */
	wrk_x = slt_tiles[t_ind].bx - cen_x;
	wrk_y = slt_tiles[t_ind].by - cen_y;
	wrk_z = slt_tiles[t_ind].bz - cen_z;
	slt_tiles[t_ind].bx = FIRST_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_x;
	slt_tiles[t_ind].by = SECOND_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_y;
	slt_tiles[t_ind].bz = THIRD_ROW(p_mat, wrk_x, wrk_y, wrk_z) + cen_z;
	if (0) {
		float wrk_length;
		int cur_type;
		
		cur_type = slt_tiles[t_ind].type;
		wrk_x = (slt_tiles[t_ind].cx - slt_tiles[t_ind].tx) 
			- my_trn_table[cur_type].cx;
		wrk_y = (slt_tiles[t_ind].cy - slt_tiles[t_ind].ty)
			- my_trn_table[cur_type].cy;
		wrk_z = (slt_tiles[t_ind].cz - slt_tiles[t_ind].tz)
			- my_trn_table[cur_type].cz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"compute_new_tile_top() : ERROR [cen - top] |tile - table_entry|= %f\n",
				wrk_length);
			exit(1);
		}
		/*
		wrk_x = (slt_tiles[t_ind].cx - slt_tiles[t_ind].tx);
		wrk_y = (slt_tiles[t_ind].cy - slt_tiles[t_ind].ty);
		wrk_z = (slt_tiles[t_ind].cz - slt_tiles[t_ind].tz);
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		fprintf(STDERR,
			"compute_new_tile_top() : DEBUG slt_tiles[%d] |cen - top|= %f\n",
			t_ind, wrk_length);
		wrk_x = (slt_tiles[t_ind].cx - slt_tiles[t_ind].bx);
		wrk_y = (slt_tiles[t_ind].cy - slt_tiles[t_ind].by);
		wrk_z = (slt_tiles[t_ind].cz - slt_tiles[t_ind].bz);
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		fprintf(STDERR,
			"compute_new_tile_top() : DEBUG slt_tiles[%d] |cen - bot|= %f\n",
			t_ind, wrk_length);
		wrk_x = my_trn_table[cur_type].cx;
		wrk_y = my_trn_table[cur_type].cy;
		wrk_z = my_trn_table[cur_type].cz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		fprintf(STDERR,
			"compute_new_tile_top() : DEBUG my_trn_table[%d] |cen - top|= %f\n",
			cur_type, wrk_length);
		wrk_x = my_trn_table[cur_type].cx - my_trn_table[cur_type].bx;
		wrk_y = my_trn_table[cur_type].cy - my_trn_table[cur_type].by;
		wrk_z = my_trn_table[cur_type].cz - my_trn_table[cur_type].bz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		fprintf(STDERR,
			"compute_new_tile_top() : DEBUG my_trn_table[%d] |cen - bot|= %f\n",
			cur_type, wrk_length);
		 */
		
		wrk_x = (slt_tiles[t_ind].cx - slt_tiles[t_ind].bx) 
			- (my_trn_table[cur_type].cx - my_trn_table[cur_type].bx);
		wrk_y = (slt_tiles[t_ind].cy - slt_tiles[t_ind].by)
			- (my_trn_table[cur_type].cy - my_trn_table[cur_type].by);
		wrk_z = (slt_tiles[t_ind].cz - slt_tiles[t_ind].bz)
			- (my_trn_table[cur_type].cz - my_trn_table[cur_type].bz);
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"compute_new_tile_top() : ERROR [cen - bot] |tile - table_entry|= %f\n",
				wrk_length);
			exit(1);
		}
		wrk_x = (slt_tiles[t_ind].bx - slt_tiles[t_ind].tx) 
			- my_trn_table[cur_type].bx;
		wrk_y = (slt_tiles[t_ind].by - slt_tiles[t_ind].ty)
			- my_trn_table[cur_type].by;
		wrk_z = (slt_tiles[t_ind].bz - slt_tiles[t_ind].tz)
			- my_trn_table[cur_type].bz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"compute_new_tile_top() : ERROR [bot - top] |tile - table_entry|= %f\n",
				wrk_length);
			exit(1);
		}
	}
	return;
}

/*
 *  Rotate the corresponding trn_table_entry.
 */
void rotate_trn_table(int t_ind, struct pe_slt_tile * slt_tiles)
{
	struct pe_matrix rot_mat;
	int cur_type;
	float wrk_x, wrk_y, wrk_z;
	float top_x, top_y, top_z;

	cur_type = slt_tiles[t_ind].type;
	if ( slt_tiles[t_ind].code == C_DOWN) {
		wrk_x = my_trn_table_save[cur_type].bx;
		wrk_y = my_trn_table_save[cur_type].by;
		wrk_z = my_trn_table_save[cur_type].bz;
	} else if ( slt_tiles[t_ind].code == C_UP) {
		wrk_x = - my_trn_table_save[cur_type].bx;
		wrk_y = - my_trn_table_save[cur_type].by;
		wrk_z = - my_trn_table_save[cur_type].bz;
	} else {
		fprintf(STDERR,
			"rotate_trn_table() : ERROR slt_tile[%d].code == UNKOWN\n", t_ind);
		exit(1);
	}
	/* rot_mat(wrk_*) == slt_tile[t_ind].d* */
	rotation_matrix(&rot_mat, wrk_x, wrk_y, wrk_z, slt_tiles[t_ind].dx,
			slt_tiles[t_ind].dy, slt_tiles[t_ind].dz, cur_type);
	
	compute_new_trn_top(&rot_mat, t_ind, slt_tiles, &top_x, &top_y, &top_z);
	compute_new_trn_downward(&rot_mat,t_ind, slt_tiles, top_x, top_y,
				 top_z);	
	compute_new_trn_upward(&rot_mat,t_ind, slt_tiles, top_x, top_y, top_z);
	
	if (0) {
		float wrk_length, wrk2_length;
		
		wrk_x = my_trn_table[cur_type].downward_cx - my_trn_table[cur_type].cx;
		wrk_y = my_trn_table[cur_type].downward_cy - my_trn_table[cur_type].cy;
		wrk_z = my_trn_table[cur_type].downward_cz - my_trn_table[cur_type].cz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
		wrk_x = my_trn_table[cur_type].upward_cx - my_trn_table[cur_type].cx;
		wrk_y = my_trn_table[cur_type].upward_cy - my_trn_table[cur_type].cy;
		wrk_z = my_trn_table[cur_type].upward_cz - my_trn_table[cur_type].cz;
		wrk2_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk2_length, 1.0)) {
			fprintf(STDERR,
				"rotate_trn_table() : ERROR my_trn_table[%d]  |downward_c* - c*|= %f \n",
				cur_type, wrk_length);
			fprintf(STDERR,
				"rotate_trn_table() : ERROR my_trn_table[%d]  |upward_c* - c*|= %f \n",
				cur_type, wrk2_length);
			exit(1);
		}
	}
	return;
}

/*
 *  Rotate the current tile.
 */
void rotate_current_tile(int t_ind, struct pe_slt_tile * slt_tiles)
{
	struct pe_matrix rot_mat;
	float wrk_x, wrk_y, wrk_z;
	
	if (0) {
		float wrk_length, wrk2_length;
		
		wrk_x = slt_tiles[t_ind].dx;
		wrk_y = slt_tiles[t_ind].dy;
		wrk_z = slt_tiles[t_ind].dz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
		if (!FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"rotate_current_tile() : ERROR |slt_tile[%d].d*|= %f\n",
				t_ind, wrk_length);
			exit(1);
		}
		wrk_x = slt_tiles[t_ind].bx - slt_tiles[t_ind].tx;
		wrk_y = slt_tiles[t_ind].by - slt_tiles[t_ind].ty;
		wrk_z = slt_tiles[t_ind].bz - slt_tiles[t_ind].tz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
		if (!FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"rotate_current_tile() : ERROR |slt_tile[%d].[bottom - top]|= %f\n",
				t_ind, wrk_length);
			if (0) print_trn_table();
			exit(1);
		}
		
		wrk_x = (slt_tiles[t_ind].cx - slt_tiles[t_ind].tx);
		wrk_y = (slt_tiles[t_ind].cy - slt_tiles[t_ind].ty);
		wrk_z = (slt_tiles[t_ind].cz - slt_tiles[t_ind].tz);
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		wrk_x = (slt_tiles[t_ind].cx - slt_tiles[t_ind].bx);
		wrk_y = (slt_tiles[t_ind].cy - slt_tiles[t_ind].by);
		wrk_z = (slt_tiles[t_ind].cz - slt_tiles[t_ind].bz);
		wrk2_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk2_length, 1.0)) {
			fprintf(STDERR,
				"rotate_current_tile() : DEBUG slt_tiles[%d] |cen - top|= %f (before)\n",
				t_ind, wrk_length);
			fprintf(STDERR,
				"rotate_current_tile() : DEBUG slt_tiles[%d] |cen - bot|= %f (before)\n",
				t_ind, wrk2_length);
			exit(1);
		}
	}
	
	/* rot_mat(current tile) == (corresponding trn_table entry) */
	rotation_matrix2(&rot_mat, t_ind, slt_tiles);
	compute_new_tile_top(&rot_mat, t_ind, slt_tiles);
	if (0) {
		float wrk_length, wrk2_length;
		
		wrk_x = (slt_tiles[t_ind].cx - slt_tiles[t_ind].tx);
		wrk_y = (slt_tiles[t_ind].cy - slt_tiles[t_ind].ty);
		wrk_z = (slt_tiles[t_ind].cz - slt_tiles[t_ind].tz);
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		wrk_x = (slt_tiles[t_ind].cx - slt_tiles[t_ind].bx);
		wrk_y = (slt_tiles[t_ind].cy - slt_tiles[t_ind].by);
		wrk_z = (slt_tiles[t_ind].cz - slt_tiles[t_ind].bz);
		wrk2_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				   + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk2_length, 1.0)) {
			fprintf(STDERR,
				"rotate_current_tile() : DEBUG slt_tiles[%d] |cen - top|= %f (after)\n",
				t_ind, wrk_length);
			fprintf(STDERR,
				"rotate_current_tile() : DEBUG slt_tiles[%d] |cen - bot|= %f (after)\n",
				t_ind, wrk2_length);
			exit(1);
		}
	}
	return;
}

void rotate_tile_and_trn_table(int t_ind, struct pe_slt_tile * slt_tiles)
{
	float wrk_length;
	
	wrk_length = sqrt(slt_tiles[t_ind].dx * slt_tiles[t_ind].dx
			  + slt_tiles[t_ind].dy * slt_tiles[t_ind].dy
			  + slt_tiles[t_ind].dz * slt_tiles[t_ind].dz);
	if (!FLOAT_NEQ(wrk_length, 1.0)) {
		if (DEBUG) fprintf(STDERR,
			"rotate_tile_and_trn_table() : DEBUG  assign new value to slt_tiles[%d].d* (= (%8.3f, %8.3f, %8.3f))\n",
			t_ind, slt_tiles[t_ind].dx, slt_tiles[t_ind].dy,
			slt_tiles[t_ind].dz);
		if (0) print_tile(t_ind, slt_tiles);
		if (slt_tiles[t_ind].code == C_UP) {
			slt_tiles[t_ind].dx = slt_tiles[t_ind].tx - slt_tiles[t_ind].bx;
			slt_tiles[t_ind].dy = slt_tiles[t_ind].ty - slt_tiles[t_ind].by;
			slt_tiles[t_ind].dz = slt_tiles[t_ind].tz - slt_tiles[t_ind].bz;
		} else {
			slt_tiles[t_ind].dx = slt_tiles[t_ind].bx - slt_tiles[t_ind].tx;
			slt_tiles[t_ind].dy = slt_tiles[t_ind].by - slt_tiles[t_ind].ty;
			slt_tiles[t_ind].dz = slt_tiles[t_ind].bz - slt_tiles[t_ind].tz;
		}
		return;
	}
	rotate_trn_table(t_ind, slt_tiles);
	rotate_current_tile(t_ind, slt_tiles);
	
	if (DEBUG) {
		int cur_type;
		float wrk_x, wrk_y, wrk_z;
		float wrk_length;
		
		cur_type = slt_tiles[t_ind].type;
		wrk_x = (slt_tiles[t_ind].cx - slt_tiles[t_ind].tx) - my_trn_table[cur_type].cx;
		wrk_y = (slt_tiles[t_ind].cy - slt_tiles[t_ind].ty) - my_trn_table[cur_type].cy;
		wrk_z = (slt_tiles[t_ind].cz - slt_tiles[t_ind].tz) - my_trn_table[cur_type].cz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"rotate_tile_and_trn_table() : ERROR |slt_tile[%d].(c* - t*) - my_trn_table[%d].c*| = %8.3f\n",
				t_ind, cur_type, wrk_length);
			if (0) print_trn_table();
			exit(1);
		}
		wrk_x = (slt_tiles[t_ind].bx - slt_tiles[t_ind].tx) - my_trn_table[cur_type].bx;
		wrk_y = (slt_tiles[t_ind].by - slt_tiles[t_ind].ty) - my_trn_table[cur_type].by;
		wrk_z = (slt_tiles[t_ind].bz - slt_tiles[t_ind].tz) - my_trn_table[cur_type].bz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"rotate_tile_and_trn_table() : ERROR |slt_tile[%d].(b* - t*) - my_trn_table[%d].b*| = %8.3f\n",
				t_ind, cur_type, wrk_length);
			if (0) print_trn_table();
			exit(1);
		}
	}
	return;
}

void move_tile_to_target(int t_ind, struct pe_slt_tile * slt_tiles, int atm_ct)
{
	float dist_x, dist_y, dist_z;
	
	dist_x = my_protein.atom[atm_ct].x - slt_tiles[t_ind].cx;
	dist_y = my_protein.atom[atm_ct].y - slt_tiles[t_ind].cy;
	dist_z = my_protein.atom[atm_ct].z - slt_tiles[t_ind].cz;
	slt_tiles[t_ind].cx = slt_tiles[t_ind].cx + dist_x;
	slt_tiles[t_ind].cy = slt_tiles[t_ind].cy + dist_y;
	slt_tiles[t_ind].cz = slt_tiles[t_ind].cz + dist_z;
	slt_tiles[t_ind].tx = slt_tiles[t_ind].tx + dist_x;
	slt_tiles[t_ind].ty = slt_tiles[t_ind].ty + dist_y;
	slt_tiles[t_ind].tz = slt_tiles[t_ind].tz + dist_z;
	slt_tiles[t_ind].bx = slt_tiles[t_ind].bx + dist_x;
	slt_tiles[t_ind].by = slt_tiles[t_ind].by + dist_y;
	slt_tiles[t_ind].bz = slt_tiles[t_ind].bz + dist_z;
	slt_tiles[t_ind].error = sqrt(dist_x * dist_x + dist_y * dist_y
		+ dist_z * dist_z);
	slt_tiles[t_ind].error_atom = atm_ct;
	return;
}
